<?php

namespace EventPortal\Domain\Form;

use Closure;
use EventPortal\Domain\Agenda\SessionErrors;
use EventPortal\Domain\Component\Input;
use EventPortal\Domain\Component\Section;
use EventPortal\Domain\Util\RelationshipType;
use EventPortal\Domain\Util\ValidationRules;
use EventPortal\Exceptions\NoSaveHandlers;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Pipeline\Pipeline;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Validator;
use Illuminate\Validation\ValidationException;

class Form
{
    /** @var string */
    public $name;

    /** @var array */
    public $blocks = [];

    /** @var array */
    public $actions = [];

    /** @var null|callable */
    public $primaryAction;

    /** @var Model */
    private $model;

    /** @var array */
    private $context = [];

    /** @var array */
    private $saveHandlers;

    /** @var bool */
    private $useAlternativeName = false;

    /** @var Validator */
    private $validator;

    public function __construct(string $name)
    {
        $this->name = $name;
    }

    /**
     * @param callable|string $fields
     */
    public function addFields($fields): self
    {
        $this->blocks[] = $fields;

        return $this;
    }

    public function addActions(callable $actions): self
    {
        $this->actions = $actions;

        return $this;
    }

    public function addPrimaryAction(callable $primaryAction): self
    {
        $this->primaryAction = $primaryAction;

        return $this;
    }

    /**
     * @param mixed ...$models
     */
    public function context(...$models): self
    {
        $keyedModels = [];

        foreach ($models as $model) {
            $keyedModels[$model::KEY] = $model;
        }

        $this->context = $keyedModels;

        if ($this->model) {
            $this->context = array_merge($this->context, [$this->model::KEY => $this->model]);
        }

        return $this;
    }

    public function model(Model $model): self
    {
        $this->model = $model;

        if ($this->context) {
            $this->context = array_merge($this->context, [$this->model::KEY => $this->model]);
        }

        return $this;
    }

    public function validate(array $input): self
    {
        $this->resolveWithoutHydration();

        $this->validator = Validator::make($input, ValidationRules::for($this->blocks));

        if ($this->useAlternativeName) {
            return $this->validateUsingAlternativeNames();
        }

        $this->validator->validate();

        return $this;
    }

    public function render(): array
    {
        $this->resolve();

        $data = [
            'actions' => $this->actions,
            'content' => $this->blocks,
        ];

        if ($this->primaryAction !== null) {
            $data['onEnter'] = $this->primaryAction;
        }

        return $data;
    }

    public function addHandler(callable $saveHandler): self
    {
        $this->saveHandlers[] = $saveHandler;

        return $this;
    }

    public function useAlternativeNames(): self
    {
        $this->useAlternativeName = true;

        return $this;
    }

    /**
     * @return mixed
     */
    public function handle(array $input)
    {
        throw_if($this->saveHandlers === [], NoSaveHandlers::class);

        $payload = Payload::make($input)
            ->withModel($this->model)
            ->withContext($this->context);

        return app(Pipeline::class)
            ->send($payload)
            ->through($this->saveHandlers)
            ->thenReturn();
    }

    private function resolveWithoutHydration(): void
    {
        $this->resolve(false);
    }

    private function resolve(bool $hydrate = true): void
    {
        $this->resolveActions();
        $this->resolveSubmit();
        $this->resolveBlocks($hydrate);
    }

    private function resolveActions(): void
    {
        if ($this->actions === []) {
            return;
        }

        $this->actions = app()->call($this->actions, $this->getContext());
    }

    private function resolveSubmit(): void
    {
        if ($this->primaryAction instanceof Closure === false) {
            return;
        }

        $this->primaryAction = app()->call($this->primaryAction, $this->getContext());
    }

    /**
     * Some documentation.
     */
    private function resolveBlocks(bool $hydrate): void
    {
        $this->blocks = collect($this->blocks)
            ->map(function ($block) {
                if ($block instanceof Closure) {
                    return app()->call($block, $this->getContext());
                }

                return $block;
            })
            ->flatten()
            ->map(function ($section) {
                if ($section instanceof Section) {
                    $section->children = $section->flatChildren();

                    return $section;
                }

                return Section::make()->fields([$section]);
            })
            ->when($hydrate, function (Collection $sections) {
                return $sections->each(function (Section $section): void {
                    foreach ($section->flatChildren() as $input) {
                        if ($this->useAlternativeName) {
                            $input->useAlternativeName();
                        }

                        $input->value($this->getValue($input));
                    }
                });
            })
            ->toArray();
    }

    /**
     * @return mixed|null
     */
    private function getValue(Input $input)
    {
        if ($input->value !== null) {
            return $input->value;
        }

        $name = $input->name;

        if ($input->isStatus()) {
            $name = 'readable_status';
        }

        if ($input->isNotScoped()) {
            return optional($this->model)->$name;
        }

        if ($this->model && RelationshipType::make($this->model, $input->scope)->isBelongsToMany()) {
            $data = $this->model->{$input->scope}->first();

            return $data->pivot->$name;
        }

        return object_get($this->model, $input->scopeKey());
    }

    public function getValidator(): \Illuminate\Validation\Validator
    {
        return $this->validator;
    }

    private function validateUsingAlternativeNames(): self
    {
        try {
            $this->validator->validate();
        } catch (ValidationException $exception) {
            $errors = SessionErrors::swapDatabaseKeysForBrowserKeys($exception->errors());

            foreach ($errors as $key => $value) {
                $this->validator->errors()->add($key, $value);
            }

            throw new ValidationException($this->validator);
        }

        return $this;
    }

    private function getContext(): array
    {
        $parameters = $this->context;

        if ($this->model) {
            return array_merge($this->context, [$this->model::KEY => $this->model]);
        }

        return $parameters;
    }
}
